; This is a record package using structural subclassing.
;
; A record class can be thought of as a set of names for the fields
; in a record of that class.
; A record class A is a subclass of a record class B if the set of
; field names associated with A is a superset of the field names
; associated with B.  If A is a subclass of B, then any instance of
; A may be used where an instance of B may be used.

; The following syntax may appear anywhere a definition may appear:
;
;   (define-record-class <classname> <fieldname1> <fieldname2> ...)
;
; The class and the field names should be symbols.
; This syntax defines the following procedures and gives them the
; same scope that a definition would have:
;
;   make-<classname>                constructs instances of the class
;   <classname>?                    true of instances of the class
;   <classname>-<fieldname1>        return the value of the named field
;   <classname>-<fieldname2>
;   ...
;   <class-name>-<fieldname1>-set!  change the value of the named field
;   <class-name>-<fieldname2>-set!
;   ...
;
; The make-<classname> procedure takes a variable but even number of
; arguments.  The first argument is a field name (a symbol) and the
; second argument is the initial value of that field.  The third
; argument is a field name, and the fourth argument is the initial
; value of that field.  Et cetera.  Fields whose initial values are
; not specified are initialized to an unspecified value.
;
; Example:
;
; >>> (define-record-class position x y z)
; position?
; >>> (define-record-class einstein x y z t)
; einstein?
; >>> (define (newtonian-distance p1 p2)
;       (define (square x) (* x x))
;       (sqrt (+ (square (- (position-x p1) (position-x p2)))
;                (square (- (position-y p1) (position-y p2)))
;                (square (- (position-z p1) (position-z p2))))))
; newtonian-distance
; >>> (define (einsteinian-distance p1 p2)
;       (define (square x) (* x x))
;       (let ((a (- (+ (square (- (einstein-x p1) (einstein-x p2)))
;                      (square (- (einstein-y p1) (einstein-y p2)))
;                      (square (- (einstein-z p1) (einstein-z p2))))
;                   (square (- (einstein-t p1) (einstein-t p2))))))
;         (if (negative? a)
;             (- (sqrt (abs a)))
;             (sqrt a))))
; einsteinian-distance
; >>> (newtonian-distance (make-position 'x 0 'y 0 'z 0)
;                         (make-position 'x 2 'y 3 'z 6))
; 7.0
; >>> (define p1 (make-einstein 'x 8 'y 4 'z 10 't 13))
; p1
; >>> (define p2 (make-einstein 'x 0 'y 0 'z 2 't 0))
; p2
; >>> (newtonian-distance p1 p2)
; 12.0
; >>> (einsteinian-distance p1 p2)
; -5.0
;
; Please note that this record package is not compatible with
; the record package proposed by Jonathan Rees, which uses name
; equivalence and does not support subtyping.

(define make-record-constructor)
(define make-record-predicate)
(define make-record-accessor)
(define make-record-setter)

; Records are represented as vector-like structures with a
; special code in element 0 and an association list of field
; names and offsets in element 1.
;
; This code is obviously non-portable, but something similar
; should be possible in any implementation of Scheme.  At worst,
; records could be represented as vectors with a special value
; in element 0.

(letrec ((vector-typetag 80)
         (record-typetag 83)
         (record-subtag 1)
         (structural-subtype?
          (lambda (fields fields-offsets)
            (if (null? fields)
                #t
                (and (assq (car fields) fields-offsets)
                     (structural-subtype? (cdr fields) fields-offsets)))))
         (construct-error
          (lambda (recordname field)
            (declare (optimize 0))
            (cerror "Illegal field name passed to record constructor" recordname field)))
         (access-error
          (lambda (recordname field x continue)
            (declare (optimize 0))
            (cerror "Illegal record field access" recordname field x)
            (continue x)))
         (set-error
          (lambda (recordname field v x continue)
            (declare (optimize 0))
            (cerror "Illegal record field assignment" recordname field v x)
            (continue x v))))
  (set! make-record-constructor
        (lambda (fields-offsets fields recordname)
          (let ((size (+ 2 (length fields))))
            (rec f
                 (lambda args
                   (let ((v (make-vector size #!unspecified)))
                     (vector-set! v 0 record-subtag)
                     (vector-set! v 1 fields-offsets)
                     (do ((args args (cddr args)))
                         ((null? args) (typetag-set! v record-typetag))
                         (let ((e (assq (car args) fields-offsets)))
                           (if e
                               (vector-set! v (cdr e) (cadr args))
                               (construct-error recordname (car args) f))))))))))
  (set! make-record-predicate
        (lambda (fields-offsets fields)
          (lambda (x)
            (and (eq? (typetag x) record-typetag)
                 (let ((y (typetag-set! x vector-typetag)))
                   (and (eq? (vector-ref y 0) record-subtag)
                        (let ((z (vector-ref y 1)))
                          (or (eq? z fields-offsets)
                              (structural-subtype? fields z)))))))))
  (set! make-record-accessor
        (lambda (fields-offsets fields field recordname)
          (let ((usual-offset (cdr (assq field fields-offsets))))
            (rec f
                 (lambda (x)
                   (if (eq? (typetag x) record-typetag)
                       (let ((y (typetag-set! x vector-typetag)))
                         (if (eq? (vector-ref y 0) record-subtag)
                             (let ((z (vector-ref y 1)))
                               (if (eq? z fields-offsets)
                                   (vector-ref y usual-offset)
                                   (if (structural-subtype? fields z)
                                       (vector-ref y (cdr (assq field z)))
                                       (access-error recordname field x f))))
                             (access-error recordname field x f)))
                       (access-error recordname field x f)))))))
  (set! make-record-setter
        (lambda (fields-offsets fields field recordname)
          (let ((usual-offset (cdr (assq field fields-offsets))))
            (rec f
                 (lambda (x v)
                   (if (eq? (typetag x) record-typetag)
                       (let ((y (typetag-set! x vector-typetag)))
                         (if (eq? (vector-ref y 0) record-subtag)
                             (let ((z (vector-ref y 1)))
                               (if (eq? z fields-offsets)
                                   (begin (vector-set! y usual-offset v) #!unspecified)
                                   (if (structural-subtype? fields z)
                                       (begin (vector-set! y (cdr (assq field z)) v)
                                              #!unspecified)
                                       (set-error recordname field x v f))))
                             (set-error recordname field x v f)))
                       (set-error recordname field x v f)))))))
  #t)



(macro define-record-class
       (lambda (l)
         (let ((name (cadr l))
               (fields (cddr l)))
           (let ((fields-offsets (do ((fields fields (cdr fields))
                                      (o 2 (1+ o))
                                      (fields-offsets '() (cons (cons (car fields) o)
                                                                fields-offsets)))
                                     ((null? fields) fields-offsets)))
                 (namestring (symbol->string name))
                 (make-name
                  (lambda args
                    (string->symbol (apply string-append args)))))
             `(begin (define ,(make-name "make-" namestring)
                       (make-record-constructor ',fields-offsets ',fields ',name))
                     ,@(map (lambda (fieldname)
                              `(define ,(make-name namestring "-" (symbol->string fieldname))
                                       (make-record-accessor ',fields-offsets
                                                             ',fields
                                                             ',fieldname
                                                             ',name)))
                            fields)
                     ,@(map (lambda (fieldname)
                              `(define ,(make-name namestring
                                                   "-"
                                                   (symbol->string fieldname)
                                                   "-set!")
                                       (make-record-setter ',fields-offsets
                                                           ',fields
                                                           ',fieldname
                                                           ',name)))
                            fields)
                     (define ,(make-name namestring "?")
                       (make-record-predicate ',fields-offsets ',fields)))))))